
module OLS

contains
  !#######################################################################
  subroutine ols_reg(n,p,x,y,beta,error_code)
    !     ### calculates OLS regression coefficients.
    !     ### Sets error_code=1 if x is rank-deficient.
    implicit none
    integer  n,p,error_code
    double precision    x(n,p),y(n),beta(p),wkp(p),wkpp1(p,p), wkpp2(p,p)
    integer  j,k
    double precision    sum
    call get_xtx(n,p,x,wkpp1)
    call get_xty(n,p,x,y,wkp)
    !     ### put inverse of wkpp1 into wkpp2 and symmetrize
    call chfce(p,p,wkpp1,error_code)
    if(error_code.eq.1) goto 999
    call bkslv(p,p,wkpp1)
    call mm(p,p,wkpp1,wkpp2)
    do j=1,p
       do k=j+1,p
          wkpp2(k,j)=wkpp2(j,k)
       end do
    end do
    !     ### multiply wkpp2 by wkp, put answer in beta
    do j=1,p
       sum=0.0D0
       do k=1,p
          sum=sum+wkpp2(j,k)*wkp(k)
       end do
       beta(j)=sum
    end do
999 continue
  end subroutine ols_reg
  !#######################################################################

  subroutine getr2(n,p,x,y,beta,r2)
    implicit none
    integer  n,p
    double precision    x(n,p),y(n),beta(p),e(n)
    integer  j,k
    double precision    ee, yy, ybar, r2

    e=y-matmul(x,beta)
	ee=sum(e**2.)
	ybar=sum(y)/n
	yy=sum((y-ybar)**2.)
	R2=1.0d0-ee/yy
  end subroutine getr2

  !#######################################################################


  subroutine get_xtx(n,p,x,xtx)
    !     ### accumulates xtx (upper-triangle part only)
    implicit none
    integer  n,p
    double precision    x(n,p),xtx(p,p)
    integer  i,j,k
    do j=1,p
       do k=j,p
          xtx(j,k)=0.0D0
       end do
    end do
    do i=1,n
       do j=1,p
          do k=j,p
             xtx(j,k)=xtx(j,k)+x(i,j)*x(i,k)
          end do
       end do
    end do
  end subroutine get_xtx
  !#######################################################################

  subroutine get_xty(n,p,x,y,xty)
    !     ###accumulates xty
    implicit none
    integer  n,p
    double precision    x(n,p), y(n), xty(p)
    integer  i,j
    do j=1,p
       xty(j)=0.0D0
    end do
    do i=1,n
       do j=1,p
          xty(j)=xty(j)+x(i,j)*y(i)
       end do
    end do
  end subroutine get_xty
  !#######################################################################


  subroutine chfce(p,pw,s,err)
    !     ### Overwrites s (upper tri) with cholesky factor
    !     ### If s is not positive definite, err is set to 1
    implicit none
    integer  p,pw,err,i,j,k
    double precision    s(p,p),sum
    err=0
    do i=1,pw
       sum=dble(0.)
       do k=1,i-1
          sum=sum+s(k,i)**2
       end do
       if(sum.ge.s(i,i)) then
          err=1
          goto 999
       endif
       s(i,i)=dsqrt(s(i,i)-sum)
       do j=i+1,pw
          sum=dble(0.)
          do k=1,i-1
             sum=sum+s(k,i)*s(k,j)
          end do
          s(i,j)=(s(i,j)-sum)/s(i,i)
       end do
    end do
999 continue
  end subroutine chfce

  !#######################################################################

  subroutine bkslv(p,pw,s)
    implicit none
    integer  p,pw,j,k,i
    double precision    s(p,p),sum
    s(1,1)=1.0D0/s(1,1)
    do k=2,pw
       s(k,k)=dble(1.)/s(k,k)
       do j=1,k-1
          sum=0.0D0
          do i=j,k-1
             sum=sum+s(j,i)*s(i,k)
          end do
          s(j,k)=-sum*s(k,k)
       end do
    end do
  end subroutine bkslv
  !#######################################################################

  subroutine mm(p,pw,wm,cm)
    !     ### calculates upper tri part of cm = wm%*%t(wm), where
    !     ### wm is upper tri
    implicit none
    integer  p,pw,i,j,k
    double precision    wm(p,p),cm(p,p),sum
    do i=1,pw
       do j=i,pw
          sum=0.0D0
          do k=max(i,j),pw
             sum=sum+wm(i,k)*wm(j,k)
          end do
          cm(i,j)=sum
       end do
    end do
  end subroutine mm
  !#######################################################################


end module OLS
