
module parameters

  implicit none
  integer:: ne, na, nsamp, ntime, nskip
  integer,parameter:: in=10
  integer,dimension(:),allocatable:: rhoopt
  real(8),dimension(:),allocatable:: agrid, egrid, mu, rho
  real(8), dimension(:,:),allocatable:: pi
  real(8):: beta, delta, AA, theta, sig, aproxdr, amin, amax
  real(8):: r, w
  real(8),parameter:: tinies=1.0d-10,zero=0.0, one=1.0
  integer:: id=0, out=10

end module parameters


module economy

  use parameters
  implicit none
  
contains

  !-------------------------------------------------------
  ! decision_rules
  !
  ! given a distribution x=(sampe,sampa),
  ! solve a maxmization problem to get ystar
  !-------------------------------------------------------
  subroutine decision_rules(sampa, sampe, ystar)
    
    implicit none
    
    integer:: ie, ia, iaa, isamp, niter
    integer,intent(in)::sampe(nsamp)
    real(8),intent(in)::sampa(nsamp)
    integer:: no_conv, no_conv_y(ne,na)
    real(8):: y0(ne,na), ystar(ne,na), tempe(nsamp)
    real(8):: aggK, aggL, income
    real(8):: update, error, xtemp(nsamp), params(4)
    logical:: dr_conv

    tempe=0.
    update=0.5d0
    niter=0
    ystar=0.

    ! given x (distribution),
    ! get aggK,varK,aggL to get r, w
    do ie=1,ne
       where(sampe==ie) tempe=egrid(ie)
    end do
       
    aggK = sum(sampa)/dble(nsamp)
    aggL = sum(tempe)/dble(nsamp)
    
    r = fk(aggK,aggL) + one - delta
    w = fl(aggK,aggL)
    
    ! initialze y0
    do ie=1,ne
       do ia=1,na
          y0(ie,ia) = r*agrid(ia) + w*egrid(ie)
       end do
    end do
    
    dr_conv = .false.
    
    do while (.not. dr_conv)
       niter=niter+1
       ystar=0.
       ! STEP 1: calculate moments for next period
       do isamp=1,nsamp
          do ia=2,na
             if (agrid(ia)>sampa(isamp)) exit
          end do
          if (ia==na+1) then
             xtemp(isamp) = y0(sampe(isamp),na) &
                  +(sampa(isamp)-agrid(na))&
                  *(y0(sampe(isamp),na)-y0(sampe(isamp),na-1))&
                  /(agrid(na)-agrid(na-1))
          else
             xtemp(isamp) = interpol(agrid(ia-1),agrid(ia),&
                  y0(sampe(isamp),ia-1),y0(sampe(isamp),ia),&
                  sampa(isamp))
          end if
       end do
       
       params(2) = sum(xtemp)/dble(nsamp)
       params(3) = sum(xtemp**2.)/dble(nsamp)
       params(4) = sum(xtemp**3.)/dble(nsamp)
       
       ! STEP 2: solve maxmization problem
       do ie=1,ne
          do ia=1,na
             income = r*agrid(ia)+w*egrid(ie)
             params(1)=income
             do iaa=1,100
             end do
             if (foc(amin,params)>zero) then
                ystar(ie,ia)=amin
             elseif (foc(income,params)<zero) then
                ystar(ie,ia)=income
             else
                do iaa=2,na
                   if (foc(agrid(iaa),params)>zero) exit
                end do
                if (iaa==na+1) then
                   ystar(ie,ia)=zbrent(agrid(na),income,aproxdr,params)
                else
                   ystar(ie,ia)=zbrent(agrid(iaa-1),agrid(iaa),aproxdr,params)
                end if
             end if
          end do
       end do
       
       ! check Convergence
       error=0.
       where (abs(ystar-y0)>aproxdr)
          no_conv_y=1
       elsewhere
          no_conv_y=0
       end where
       
       no_conv = sum(no_conv_y)
       error=maxval(abs(ystar-y0))
       
       if (no_conv==0) then
          dr_conv=.true.
       else
          dr_conv=.false.
       end if
       
       if (dr_conv==.false.) then
          y0 = update*ystar + (one-update)*y0
       end if
       
    end do
  end subroutine
  

  real(8) function foc(yy,params)
    implicit none
    real(8),intent(in)::yy,params(4)
    real(8)::income
    
    income=params(1)
    foc=Uc(income-yy)-beta*dVfcn(params,yy)
    
  end function foc


  real(8) function dVfcn(params,yy)
    implicit none
    integer::ii
    real(8),intent(in)::params(4), yy
    real(8):: dV
    
    dvfcn=0.
    
    ! V(K1,K2) =  rho1*log(K1) + rho2*log(K2) +rho3*log(K3) + rho4
    
    dvfcn = &
         rho(1)/params(2) + &
         rho(2)/params(3)*2.*yy + &
         rho(3)/params(4)*3.*yy**2. + &
         rho(4)/params(2)*2.*dlog(params(2)) + &
         rho(5)/params(3)*2.*dlog(params(3))*2.*yy +&
         rho(6)/params(4)*2.*dlog(params(4))*3.*yy**2. +&
         rho(7)/params(2)/params(3)*&
         (params(2)*dlog(params(2))*2.*yy + &
         params(3)*dlog(params(3))) +&
         rho(8)/params(2)/params(4)*&
         (params(2)*dlog(params(2))*3.*yy**2. + &
         params(4)*dlog(params(4))) +&
         rho(9)/params(3)/params(4)*&
         (params(3)*dlog(params(3))*3.*yy**2. + &
         params(4)*dlog(params(4))*2.*yy)
    
  end function dVfcn


  FUNCTION Fk(kin,lin)
    implicit none
    real(8):: kin,lin,Fk
    
    Fk=AA*theta*kin**(theta-1.)*lin**(1.-theta)
    
  END FUNCTION Fk
  
  FUNCTION Fl(kin,lin)
    implicit none
    real(8):: kin,lin,Fl
    
    Fl=AA*(1.-theta)*kin**theta*lin**(-theta)
    
  END FUNCTION Fl
  
  
  FUNCTION INTERPOL (a0,a1,y0,y1,ax) RESULT (yint)
    implicit none
    real(8):: a0, a1, y0, y1, ax, yint
    yint=y0+(ax-a0)/(a1-a0)*(y1-y0)
  END FUNCTION INTERPOL
  
  FUNCTION Uc(cin)
    implicit none
    real(8):: cin,Uc
    
    if (cin<tinies) then
       Uc=500./tinies
    else
       Uc=cin**(-sig)
    end if
    
  END FUNCTION Uc

  !********************************************************************
  ! function: zbrent
  ! from Numerical Recipes in Fortran 77, pp354-355.
  !********************************************************************
  FUNCTION zbrent (x1,x2,tol,params)
    implicit none
    real(8), intent(in):: x1,x2,tol
    integer, parameter:: itmax=3000
    real(8), parameter:: epsz=3.e-8
    integer:: iter
    real(8):: a,b,c,d,e,fa,fb,fc,p,q,r,s,tol1,xm, zbrent,params(4)

    a=x1
    b=x2
    fa=foc(a,params)
    fb=foc(b,params)
    if(fa*fb>0.) pause 'root must be bracketed for zbrent'
    c=b
    fc=fb
    do iter=1,itmax
       if (fb*fc>0.) then
          c=a
          fc=fa
          d=b-a
          e=d
       end if
       if (abs(fc)<abs(fb)) then
          a=b
          b=c
          c=a
          fa=fb
          fb=fc
          fc=fa
       end if
       tol1=2.*epsz*abs(b)+0.5*tol
       xm=.5*(c-b)
       if (abs(xm)<=tol1 .or. fb==0.) then
          zbrent=b
          return
       end if
       if (abs(e)>=tol1 .and. abs(fa)>abs(fb)) then
          s=fb/fa
          if(a==c) then
             p=2.*xm*s
             q=1.-s
          else
             q=fa/fc
             r=fb/fc
             p=s*(2.*xm*q*(q-r)-(b-a)*(r-1.))
             q=(q-1.)*(r-1.)*(s-1.)
          end if
          if(p>0.) q=-q
          p=abs(p)
          if (2.*p<min(3.*xm*q-abs(tol1*q),abs(e*q))) then
             e=d
             d=p/q
          else
             d=xm
             e=d
          end if
       else
          d=xm
          e=d
       end if
       a=b
       fa=fb
       if(abs(d)>tol1) then
          b=b+d
       else
          b=b+sign(tol1,xm)
       end if
       fb=foc(b,params)
    end do
    pause 'zbrent exceeding maximum iterations'
    zbrent=b
    return
  END FUNCTION zbrent

  subroutine generate_sample(sampa, sampe, ystar, kk, k4, rr)
    implicit none
    
    integer:: ie, ia, iaa, isamp
    integer::sampe(nsamp), sampep(nsamp)
    real(8)::sampa(nsamp), sampap(nsamp), eshock(nsamp)
    real(8)::sampc(nsamp), sampu(nsamp)
    integer:: no_conv, no_conv_y(ne,na)
    real(8):: ystar(ne,na)
    real(8):: kk(3), Kkp(3), income
    real(8):: xtemp, rr,k4
    
    kk(1) = sum(sampa)/dble(nsamp)
    kk(2) = sum(sampa**2.)/dble(nsamp)
    kk(3) = sum(sampa**3.)/dble(nsamp)
    k4 = sum(sampa**4.)/dble(nsamp)
    call random_number(eshock)
    
    rr = 0.
    do isamp=1,nsamp
       do ia=2,na
          if (agrid(ia)>sampa(isamp)) exit
       end do
       if (ia==na+1) then
          xtemp = ystar(sampe(isamp),na)
       else
          xtemp = interpol(agrid(ia-1),agrid(ia),&
               ystar(sampe(isamp),ia-1),&
               ystar(sampe(isamp),ia),&
               sampa(isamp))
       end if
       income = r*sampa(isamp)+w*egrid(sampe(isamp))
       rr = rr+util(income-xtemp)
       
       sampap(isamp)=xtemp
       sampc(isamp)=income-sampap(isamp)
       sampu(isamp)=util(sampc(isamp))
       
       do ie=1,ne
          if (eshock(isamp)<sum(pi(sampe(isamp),1:ie))) then
             sampep(isamp)=ie
             exit
          end if
       end do
    end do

    rr=rr/nsamp
    
    kkp(1) = sum(sampap)/dble(nsamp)
    kkp(2) = sum(sampap**2.)/dble(nsamp)
    kkp(3) = sum(sampap**3.)/dble(nsamp)
    rr = rr + beta*Vfcn(kkp)
    
    sampa = sampap
    sampe = sampep
    
  end subroutine generate_sample

  real(8) function Vfcn(kk)
    implicit none
    integer:: ii
    real(8):: kk(3)
    
    vfcn=rho(1)*dlog(kk(1))+rho(2)*dlog(kk(2))+rho(3)*dlog(kk(3))+&
         rho(4)*(dlog(kk(1)))**2. + &
         rho(5)*(dlog(kk(2)))**2. + &
         rho(6)*(dlog(kk(3)))**2. + &
         rho(7)*dlog(kk(1))*dlog(kk(2)) + &
         rho(8)*dlog(kk(1))*dlog(kk(3)) + &
         rho(9)*dlog(kk(2))*dlog(kk(3)) + &
         rho(10)
    
  end function Vfcn
  
  real(8) function util(cin)
    implicit none
    real(8):: cin

    if (cin<tinies) then
       Util=tinies**(one-sig)/(one-sig)-tinies
    else
       Util=cin**(one-sig)/(one-sig)
    end if
    
  END FUNCTION Util

  SUBROUTINE initialize
    implicit none
    integer:: i, nn, mmax1, mmax2, m1, m2, amax1, amax2
    real(8):: amintemp, amaxtemp, temp
    real(8),dimension(:),allocatable::agrid1,agrid2,agridt,mgrid1,mgrid2,mgridt
    character(2)::expn, cid

    ! Read Parameters

    open(in,file="input.txt")

    read(in,*)
    read(in,*) expn
    read(in,*) amin
    read(in,*) amax
    read(in,*) na
    read(in,*) beta
    read(in,*) AA
    read(in,*) theta
    read(in,*) sig
    read(in,*) delta


    read(in,*) aproxdr
    read(in,*) !mmax
    read(in,*) !aproxdis
    read(in,*) !rstart, rmin, rmax

    read(in,*) ne
    allocate(egrid(ne));egrid=0.0
    read(in,*) egrid

    allocate(pi(ne,ne)); pi=0.0
    do i=1,ne
       read(in,*) pi(i,:)
    end do

    allocate(mu(ne)); mu=0.0
    do i=1,ne
       read(in,*) mu(i)
    end do
    
    read(in,*) !nc
    read(in,*) 
    read(in,*) nsamp
    read(in,*) nskip, ntime

    amax1=na/2
    amax2=na-amax1
    allocate(agrid1(amax1),agrid2(amax2),agridt(na))
    
    do nn=1,amax1
       agrid1(nn)=amin+(amax-amin)*(nn-1)/(amax1-1)
    end do
    do nn=1,amax2
       agrid2(nn)=dsqrt(amax-amin)*(nn-1)/(amax2-1)
    end do
    agrid2=amin+agrid2**2.
    agrid2(amax2)=amax
    agrid2(1)=amin
    
    m1=1;m2=1
    
    do nn=1,amax1+amax2
       if (agrid1(m1)>agrid2(m2)) then
          agridt(nn)=agrid2(m2)
          m2=m2+1
       elseif (agrid1(m1)<agrid2(m2)) then
          agridt(nn)=agrid1(m1)
          m1=m1+1
       else
          agridt(nn)=agrid1(m1)
          m1=m1+1
          m2=m2+1
       end if
       na=nn
       if (m1>amax1 .or. m2>amax2) exit
    end do
    agridt(1)=amin
    agridt(agrid)=amax
    
    
    allocate(agrid(na));agrid=0.0
    agrid=agridt(1:na)

    allocate(rho(10));rho=0.
    allocate(rhoopt(10));rhoopt=0
    !open(out,file='out.txt')
    
  end subroutine initialize

  !--------------------------------------------------------------------

  subroutine regress_LOM(kk,rr,rhot,r2)

    use OLS
    implicit none
    
    integer itime, nobs, error_code, npar, ipar, ii
    real(8) r2, rhot(10), kk(ntime,3), rr(ntime), RHS(ntime-nskip,10)
    real(8), dimension(:), allocatable:: rhotemp
    real(8), dimension(:,:), allocatable:: RHSTEMP

    nobs = ntime-nskip
    npar = sum(rhoopt)

    allocate(rhotemp(npar)); rhotemp=0.
    allocate(RHSTEMP(ntime-nskip,npar)); RHSTEMP=0.
    RHS = one
    RHS(:,1:3) = dlog(kk(nskip+1:ntime,:))
    RHS(:,4:6) = RHS(:,1:3)**2.
    RHS(:,7) = RHS(:,1)*RHS(:,2)
    RHS(:,8) = RHS(:,1)*RHS(:,3)
    RHS(:,9) = RHS(:,2)*RHS(:,3)
    
    ipar=1
    do ii=1,10
       if (rhoopt(ii)==0) cycle
       RHSTEMP(:,ipar)=RHS(:,ii)
       ipar=ipar+1
    end do
    

    call ols_reg(nobs,npar,RHSTEMP,rr(nskip+1:ntime),rhotemp,error_code)
    call getr2(nobs,npar,RHSTEMP,rr(nskip+1:ntime),rhotemp,r2)
    
    ipar=1
    do ii=1,10
       if (rhoopt(ii)==0) cycle
       rhot(ii)=rhotemp(ipar)
       ipar=ipar+1
    end do
    
  end subroutine Regress_LOM
  
  
  SUBROUTINE sampinit(sampa,sampe,ninit)
    
    real(8)::sampa(nsamp), shocks(nsamp), shocks2(nsamp), mucum(ne)
    integer::sampe(nsamp), ie, isamp, ninit
    
    sampa=0.
    sampe=0
    
    mucum=mu
    do ie=2,ne
       mucum(ie)=mucum(ie-1)+mu(ie)
    end do
    
    call random_init
    call random_number(shocks)
    call random_number(shocks2)
    

    sample_loop: do isamp=1,nsamp
       sampa(isamp)=shocks2(isamp)*3.651+(1.0d0-shocks2(isamp))*(7.651)
       do ie=1,ne
          if (shocks(isamp) .le. mucum(ie)) then
             sampe(isamp)=ie
             cycle sample_loop
          endif
       end do
    end do sample_loop
    

  end subroutine sampinit
  
  
  SUBROUTINE random_init
    implicit none
    integer:: nseed, seed0
    integer, allocatable, dimension(:):: seed_old, seed_new

    call random_seed
    call random_seed (SIZE=nseed)
    allocate (seed_old(nseed), seed_new(nseed))
    call random_seed (GET=seed_old(1:nseed))
    do seed0=1, nseed
       seed_new(seed0)=(100+id)*seed0
    end do
    seed_new=5
    call random_seed (PUT=seed_new(1:nseed))
    deallocate (seed_old, seed_new)

  END SUBROUTINE random_init


  subroutine calculate_stat(sampe, sampa, frac, st)
    implicit none
    
    integer,intent(in)::sampe(nsamp)
    real(8),intent(in)::sampa(nsamp)
    integer::hiasset(nsamp), ie
    real(8):: st(4), frac, aggL, tempe(nsamp)
    
    hiasset=0
    
    do ie=1,ne
       where(sampe==ie) tempe=egrid(ie)
    end do
    aggL = sum(tempe)/dble(nsamp)
    
    where(sampa>3.) hiasset=1
    st(1) = sum(sampa)/dble(nsamp)
    st(2) = sum((sampa-st(1))**2.)/dble(nsamp)
    st(3) = sum((sampa-st(1))**3.)/dble(nsamp)/(st(2)**1.5)
    st(4) = sum((sampa-st(1))**4.)/dble(nsamp)/(st(2)**2)
    frac = dble(sum(hiasset))/dble(nsamp)
       
    st(1) = st(1)/aggL
    st(2) = st(2)/(aggL**2.)
    
  end subroutine
     
  subroutine store_output(rho,ystar,sampa,sampe)
       
    real(8)::sampa(nsamp), ystar(ne,na), rho(3)
    integer::sampe(nsamp)
    
  end subroutine store_output

end module


program main
  use parameters
  use economy
  
  implicit none
  integer:: itime, final, LOMiter=0, final_iter=0
  real(8):: LOMerr, r2, frac, k4
  integer,dimension(:),allocatable::sampe
  real(8),dimension(:),allocatable::sampa, rr, rhot, st
  real(8),dimension(:,:),allocatable::ystar, kk
  logical:: LOM_conv=.false.
  
  call initialize
  allocate(sampe(nsamp),sampa(nsamp));sampe=0;sampa=0.
  allocate(rr(ntime));rr=0.
  allocate(ystar(ne,na));ystar=0.
  
  allocate(kk(ntime,3));kk=0.
  allocate(rhot(10));rhot=0.
  allocate(st(4));st=0.
  
  open(19,file='rhoin.txt')
  read(19,*) rho
  read(19,*) rhoopt
  close(19)
  
  open(20,file='LOMiter.txt')
  do while (.not. LOM_conv)
     
     LOMiter=LOMiter+1
     rhot=0.
     
     open(25,file="tseries.txt")
     write(25,26) "%Time series of K with rho=(", rho, ")"
26   format(A,10f9.4,A)
     write(25,*) "%   aggK     varK      kurtK      PLutil    minA     maxA"
     open(333,file="simrlt.txt")
     write(333,26) "%Time series with rho=(", rho, ")"
     
     do itime=1,ntime
        if (itime==1) call read_dist(sampe,sampa)
        call decision_rules(sampa, sampe, ystar)
        call calculate_stat(sampe,sampa,frac,st)
        call generate_sample(sampa, sampe, ystar, kk(itime,:), k4, rr(itime))
        write(25,27) itime,kk(itime,:), rr(itime), minval(sampa), maxval(sampa)
27      format(i5,6f20.5)
        if (mod(itime,20)==1) write(*,'(2x,i5,5f20.5)') itime,kk(itime,:), k4, frac
        write(333,'(2x,i5,10f20.5)') itime,kk(itime,:), k4, frac,st, rr(itime)
     end do
     close(25)
     close(333)
     print*, "regression"
     call regress_LOM(kk,rr,rhot,r2)
     
     
     
     LOMerr=maxval(dabs(rhot-rho))
     write(*,21) LOMiter,rho
     write(20,21) LOMiter,rho
     write(*,22) rhot,r2,LOMerr
     write(20,22) rhot,r2,LOMerr
21   format (i5,' rho =( ', 10f10.5,')')
22   format (5x,' rhop=( ', 10f10.5,')  R2=',f9.6,' err=',f10.5)
     
     if (LOMerr<1.0d-5 .or. final_iter==1) then
        LOM_conv=.true.
     else
        rho=0.2d0*rhot+0.8d0*rho
        rhot=0.
     end if
     
  end do
  
  close(20)
  call store_output(rho,ystar,sampa,sampe)
  
end program main

